//---------------------------------------------------------------------------
// Copyright 1998 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/IO/Smart/Buffer.h,v $
// $Id: Buffer.h,v 1.4 2005/01/16 20:28:15 ah Exp $
//
// Description: 
//
//---------------------------------------------------------------------------

#ifndef AH_IO_SMART_BUFFER_H		// prevent multiple includes
#define AH_IO_SMART_BUFFER_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
#include <AH/Error/Sys.h>
#include <AH/Error/Intern.h>
USING_NAMESPACE(AH)

#include <cerrno>
#include <string>
#ifdef BOUNDS_CHECK
#   include<iostream>
#endif
USING_NAMESPACE(std)

// --- types and classes ---

BEGIN_NAMESPACE(AH)

/// smart buffer for any type <tt>T</tt>
/**
   Use <tt>AH::IOSmartBuffer</tt> if you need a resizable buffer or
   a buffer which gets properly removed if the scope is left, e.g.
   due to an exception.

   Please include <em>AH/IO/Smart/Buffer.h</em>.
 */
template <class T>
class IOSmartBuffer
{
 public:
    /// constructor for buffer of <tt>s</tt> elements of type <tt>T</tt>
    IOSmartBuffer(size_t s)
	: size(s), max(size), buf (new T[max])
    {
	checkBuf();
    }

    /// constructor for empty buffer of type <tt>T</tt>
    IOSmartBuffer() : size(0), max(0), buf (0) {}

    /// destructor
    ~IOSmartBuffer() { if (buf) delete[] buf; }

    /// resize buffer for <tt>s</tt> elements of type <tt>T</tt>
    /**
      <tt>resize(s)</tt> resizes the buffer to contains <tt>s</tt> elements of 
      type <tt>T</tt>:
      - if <tt>s</tt> is bigger than the current size then the old buffer
        is destroyed and a new one with the appropriate size is created.
      - if <tt>s</tt> is smaller than the current size then nothing is done at all.
      - if <tt>s</tt> is equal to the current size then nothing is done at all.
      .
      This means that the buffer does only grow but never shrinks. This
      speeds up the execution but blocks memory.
     */
    void resize(size_t s)
    {
	if (s == size) return;
	if (s > max) recreate(s);
	else	     size = s;
    }

    /// resize buffer for <tt>s</tt> elements of type <tt>T</tt>
    /**
      <tt>recreate(s)</tt> resizes the buffer to contains <tt>s</tt> elements of 
      type <tt>T</tt>:
      - if <tt>s</tt> is not equal to the current size then the old buffer
            is destroyed and a new one with the appropriate size is created.
      - if <tt>s</tt> is equal to the current size then nothing is done
            at all.
      .
      This means that the buffer grows and shrinks according to the new 
      size <tt>s</tt>. This costs execution time but does not block memory.
     */
    void recreate(size_t s)
    {
	if (s == size) return;
	if (buf) delete[] buf;
	max = size = s;
	buf = new T[max];
	checkBuf();
    }

    /// remove buffer and free its allocated memory
    void remove()
    {
	if (buf) delete[] buf;
	buf = 0;
	max = size = 0;
    }

    /// clear all buffer entries (set to 0)
    void clear()
    {
	memset(buf, 0, getSize());
    }

    /// set all buffer entries to specified value
    void initAll(const T& t)
    {
	for (unsigned int i = 0; i < size; ++i)
	    buf[i] = t;
    }

    /// return number of types <tt>T</tt> which can be stored in buffer
    size_t getNum() const { return size; }

    /// return size of buffer in bytes
    size_t getSize() const { return size * sizeof(T); }

    /// return pointer to buffer
    T* operator()() const { return buf; }

    /// return specified entry in buffer (with bound check)
    const T& getEntry(size_t idx) const 
    {
#ifdef BOUNDS_CHECK
	if (idx < size) return buf[idx];
	throw ErrorInternBadArg("in IOSmartBuffer::getEntry()");
#else
	return buf[idx];
#endif
    }

    /// set specified entry in buffer (with bound check)
    void setEntry(size_t idx, const T& t) const 
    {
#ifdef BOUNDS_CHECK
	if (idx < size) buf[idx] = t;
	else throw ErrorInternBadArg("in IOSmartBuffer::setEntry()");
#else
	buf[idx] = t;
#endif
    }

    /// return specified readonly entry in buffer (no bound check)
    const T& operator[](size_t idx) const
    {
#ifdef BOUNDS_CHECK
	if (idx < size) return buf[idx];
	throw ErrorInternBadArg("in IOSmartBuffer::operator[]");
#else
	return buf[idx];
#endif
    }

    /// return specified writable entry in buffer (no bound check)
    T& operator[](size_t idx)
    {
#ifdef BOUNDS_CHECK
	if (idx < size) return buf[idx];
	throw ErrorInternBadArg("in IOSmartBuffer::operator[]");
#else
	return buf[idx];
#endif
    }

 private:
    size_t	size;	// current size of buffer in number of entries
    size_t	max;	// maximal size of buffer in number of entries
    T*		buf;	// the buffer itsself

    /// check if buffer could be created
    void checkBuf()
    {
	if (!buf) 
	{
	    max = size = 0;
	    throw ErrorSys(ENOMEM);
	}
    }
};	// IOSmartBuffer


END_NAMESPACE
#endif // !AH_IO_SMART_BUFFER_H
