//---------------------------------------------------------------------------
// Copyright 2004 Andreas Huennebeck
//
// Licensed under the terms of the GNU Library General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING.LIB' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHio/CmdOptHandler.c,v $
// $Id: 
//
// Description: implementation of class IOCmdOption
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: CmdOptHandler.c,v 1.2 2004/04/20 22:26:22 ah Exp $ (c) 2004 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h>
#include <AH/IO/CmdOptHandler.h>
#include <AH/IO/CmdOption.h>
#include <AH/Error/Sys.h>
#include "fgetopt.h"
USING_NAMESPACE(AH)

#include <cerrno>
USING_NAMESPACE(std)

// --- local stuff ---

#define TRACE 0

// --- class member implementation ---

BEGIN_NAMESPACE(AH)


void IOCmdOptHandler::addOptions(const std::vector<IOCmdOption*> newOpts)
{
    for (unsigned int i = 0; i < newOpts.size(); ++i)
	opts.push_back(newOpts[i]);
}

// return help text of all options.
std::string IOCmdOptHandler::getHelpText(const std::string& lineHead)
{
    std::string helptext;
    for (unsigned int i = 0; i < opts.size(); ++i)
    {
	helptext += lineHead;
	if (opts[i]->isShort())
	{
	    helptext += "-";
	    helptext += opts[i]->getShortOpt();
	}
	if (opts[i]->isLong())
	{
	    if (opts[i]->isShort()) helptext += ", ";
	    helptext += "--";
	    helptext += opts[i]->getLongOpt();
	}
	helptext += "\t";
	helptext += opts[i]->getDescription();
	helptext += "\n";
    }
    return helptext;
}

// return short option string, for help text.
std::string IOCmdOptHandler::getShortOptionText() const
{
    if (0 == short_opt.size()) return "";
    return "[-" + short_opt + "]"; 
}



// return next entry if this is an argument:
std::string IOCmdOptHandler::getNextOptArg(int& a, int argc, char* argv[])
{
    if (a+1 >= argc) return ""; // no more arguments available

    const char* next = argv[a+1];
    if (!next) return ""; // no argument
    if (0 == strcmp(next, "-"))
    {
	++a;
	return "-"; // argument is stdin/out
    }
    if (next[0] == '-') return ""; // this is an option, not an argument

    ++a;
    return next; // this is an argument
}



int IOCmdOptHandler::handleArgs
(
    int argc,
    char* argv[],
    std::vector<std::string>& remarg,
    std::vector<std::string>& errwarn
)
{
    //-------------------------------------------------------
    // build table of long options, and short option string:
    //-------------------------------------------------------

    int badArg = 0;
    int a = 1;
    for (; a < argc; ++a)
    {
	if (!argv[a]) 
	{
	    ++badArg;
	    continue;
	}

	std::string curArg = argv[a];
	if (0 == strncmp(curArg.c_str(), "--", 2))
	{
	    // assume long option
	    if (curArg.size() == 2)
	    {
		errwarn.push_back("incomplete long option '--'");
		continue;
	    }
	    // it is a long option: check for additional argument:
	    string::size_type pos = curArg.find_first_of("=",0);
	    if (string::npos == pos)
	    {
		optArg = getNextOptArg(a, argc, argv);
	    }
	    else
	    {
	        optArg = curArg.substr(pos));
		curArg = curArg.substr(0, pos));
	    }
	    badArg += handleLongArg(curArg.c_str() + 2, optArg, errwarn);
	}
	else if (curArg[0] == '-')
	{
	    // assume short option
	    if (curArg.size() == 1)
		break; // this is stdin or stdout, not an arg
	    // it is a short option
	    std::string optArg;
	    if (curArg.size() > 2)
	    {
		optArg = curArg.c_str() + 2;
	    }
	    else optArg = getNextOptArg(a, argc, argv);
	    badArg += handleShortArg(curArg.c_str() + 1, optArg, errwarn);
	}
    }

    // any remaining stuff are arguments:
    while (a < argc)
	remarg.push_back(argv[a++]);

    return badArg;

}



int IOCmdOptHandler::handleLongArg
(
    const char* name, 
    const std::string& optarg,
    std::vector<std::string>& errwarn
)
{
    int errors = 0;
    bool found = false;
    for (unsigned int i = 0; i < opts.size(); ++i)
    {
	if (0 == strcmp(opts[i]->getLongOpt(), name)) 
	{
	    found = true;
	    opts[i]->set();
	    switch (opts[i]->getSecArgMode())
	    {
	    default:
	    case IOCmdOption::SecondArgNone: break;
	    case IOCmdOption::SecondArgAlways:
		if (optarg == "")
		{
		    errwarn.push_back("missing argument for --" +
				      std::string(name));
		    ++errors;
		}
		// fall through
	    case IOCmdOption::SecondArgOptional:
		if (optarg != "") opts[i]->setSecArg(optarg);
		break;
	    }
	    break;
	}
    }
    if (!found)
    {
	errwarn.push_back("invalid option --" +
			  std::string(name));
	++errors;
    }
    return errors;
}


int IOCmdOptHandler::handleShortArg
(
    const char* name,
    const std::string& optarg,
    std::vector<std::string>& errwarn
)
{
    int errors = 0;
    bool found = false;
    for (unsigned int i = 0; i < opts.size(); ++i)
    {
	if (opts[i]->getShortOpt() == name[0]) 
	{
	    found = true;
	    opts[i]->set();
	    switch (opts[i]->getSecArgMode())
	    {
	    default:
	    case IOCmdOption::SecondArgNone:
		if (!name[1]) // short option contains second argument
		{
		    errwarn.push_back("bad short option -" + 
				      std::string(name));
		    ++errors;
		}
		break;
	    case IOCmdOption::SecondArgAlways:
		if (optarg == "")
		{
		    errwarn.push_back("missing argument for -" +
				      std::string(name));
		    ++errors;
		}
		// fall through
	    case IOCmdOption::SecondArgOptional:
		if (optarg != "") opts[i]->setSecArg(optarg);
		break;
	    }
	    break;
	}
    }
    if (!found)
    {
	errwarn.push_back("invalid option -" +
			  std::string(name));
	++errors;
    }
    return errors;
}


END_NAMESPACE


