//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/mpegtools/mpegcut.c,v $
// $Id: mpegcut.c,v 1.11 2004/12/31 14:11:45 ah Exp $ 
//
// Description: cut a pice out of a MPEG file into a new MPEG file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: mpegcut.c,v 1.11 2004/12/31 14:11:45 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
#include <AH/Config/Types.h>
USING_NAMESPACE(AH)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
#include <AH/IO/Mpeg/FrameHeader.h>
USING_NAMESPACE(AH)

#include "fileset.h"

// --- local constants ---

static const char* cmdLineOptions = "f:hi:t:vV";
static const char* version = "1.1";

// --- local variables ---

static int verbose = 0;

// --- local classes ---

// --- local functions ---


static void usage(bool help = false)
{
    cerr << "Usage: mpegcut [-" << cmdLineOptions << "] <output file>\n"
	 << "       cut piece(s) from MPEG file(s) into new MPEG file.\n"
	 << "  options:\n"
	 << "   -h                  - print this help message\n"
	 << "   -i <input file>     - specify input file\n"
	 << "   -f <frame-range>    - cut specified range of frames\n"
	 << "      e.g '-f 100-200'   cut frames 100 to 200 inclusively\n"
	 << "      e.g '-f 100-'      cut from frame 100 to end of file\n"
	 << "      e.g '-f -200'      cut from begin of file to frame 200\n"
	 << "   -t <time-range>     - cut specified time range\n"
	 << "                         time is given as Min:Sec.MilliSec\n"
	 << "      e.g '-t 0:12.3-3:15.23'\n"
	 << "      e.g '-t 0:12.3-'   cut from 0:12.3 to end of file\n"
	 << "      e.g '-t ..3:15.23' cut from begin of file to 3:15.23\n"
	 << "   -v                  - verbose mode (increase debug level)\n"
	 << "   -V                  - print version\n"
	 << "  arguments:\n"
	 << "   <output file>       - output file ('-' for stdout)\n"
	 << "\n"
	 << "All files must be MPEG files. The supported versions are:\n"
	 << "- MPEG-1 Layer 1, Layer 2 (MP2) and Layer 3 (MP3)\n" 
	 << "- MPEG-2 Layer 1, Layer 2 and Layer 3\n" 
	 << "- MPEG-2.5 Layer 1, Layer 2 and Layer 3\n" 
	 << "\n" 
	 << "Note: The order of the options is: -f/-t -i\n"
	 << "Bugs: The option -t is not fault tolerant for typing errors."
	 << endl;
    if (!help) return;

    cerr << "\nExamples:\n"
	 << "  The following examples assume these MPEG files as given:\n"
	 << "  - a.mpg with 5 minutes play time\n"
	 << "  - b.mpg with 4 minutes play time\n"
	 << "\n"
	 << "1. Concatenate both files a.mpg and b.mpg into file c.mpg:\n"
	 << "     mpegcut -i a.mpg -i b.mpg c.mpg\n"
	 << "   The resulting file c.mpg has 9 minutes play time.\n"
	 << "\n"
	 << "2. Extract the 2nd minute of file a.mpg into c.mpg:\n"
	 << "     mpegcut -t 1:00-2:00 -i a.mpg b.mpg\n"
	 << "   The resulting file c.mpg has 1 minute play time.\n"
	 << "\n"
	 << "3. Play the first 10 seconds of file a.mpg (use mpg123 as player):\n"
	 << "     mpegcut -t -0:10 -i a.mpg - | mpg123 -\n"
	 << "\n"
	 << "4. Extract the last minute of file a.mpg and the first minute\n"
	 << "   of b.mpg into c.mpg:\n"
	 << "     mpegcut -t 4:00- -i a.mpg -t -1:00 -i b.mpg c.mpg\n"
	 << "   The resulting file c.mpg has 2 minutes play time.\n"
	 << "\n"
	 << "5. Play the 10 seconds of the cut of example 4 (use mpg123 as player):\n"
	 << "     mpegcut -t 4:55- -i a.mpg -t -0:05 -i b.mpg - | mpg123 -\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "mpegcut Vs " << version << " " 
	 << "(c) 2000 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//----------------------------------------------------------------
// check input files for existence and for matching audio format
// On Error: throws an exception
//----------------------------------------------------------------
static void check_input_files(const FileSettings& files)
{
    for (unsigned int i=0; i < files.getNum(); i++)
    {
	const string& inputFileName = files.getFileName(i);
	int ret = access(inputFileName.c_str(), R_OK);
	if (-1 == ret)
	{
	    throw ErrorSys("Cannot access \"" + inputFileName + "\"");
	}
    }
} // end of check_input_files()


//-------------------------------------------
// check if 'val' is within specified range:
// - return -1 if val is smaller than range
// - return  0 if val is inside
// - return -1 if val is bigger than range
//-------------------------------------------
template <class T> static int inRange(T val, T min, T max)
{
    if (0 > min)
    {
	// no minimum defined: range starts from begin
	if (0 > max) return 0;
	if (val <= max) return 0;
	return 1;    // val > max
    }

    if (0 > max)
    {
	// no maximum defined: range ends at end
	if (val >= min) return 0;
	return -1;   // val < min
    }

    if (val >= min)
    {
	if (val <= max) return 0;
	return 1;    // val > max
    }
    return -1;       // val < min
}


//----------------------------------------------------------------
// read input file and write specified part into output file
// On Error: throws an exception
//----------------------------------------------------------------
static void handle_input_file
(
    const FileSettings& files,     // what to do with the input files
    unsigned int i,                // index to current input file
    const string& outputFileName,  // name of output file
    FILE* out,                     // output file stream
    unsigned int& frameOutNum      // number of frames written
)
{
    const string& inputFileName = files.getFileName(i);

    // get range (frames and times):
    int f1, f2;
    files.getFrameRange(i, f1, f2);
    double t1, t2;
    files.getTimeRange(i, t1, t2);

    bool useTime = false;
    if ((0 <= t1) || (0 <= t2)) useTime = true;

    // open input file:
    if (verbose > 1) cerr << "open file: " << inputFileName << endl;
    FILE* fp = fopen(inputFileName.c_str(), "r");
    if (!fp) return;

    // read from input file:
    if (verbose > 2) cerr << "reading file: " << inputFileName << endl;
    int frameNum = 0;
    unsigned int pos = 0;
    unsigned int badBytes = 0;
    unsigned int badFrames = 0;
    double acctime = 0;
    while(1)
    {
	uint8 header[4]; 

	//--------------------
	// read frame header:
	//--------------------
	int ret = fread(&header[0], sizeof(header[0]), 1, fp);
	if (ret != 1) break; // EOF reached
	if (verbose > 3) cerr << "byte[" << pos << "]=" 
			      << (unsigned int)header[0] << endl;
	pos++;
	if (!IOMpegFrameHeader::isFirstByte(header[0]))
	{
	    badBytes++;
	    continue;
	}
	
	ret = fread(&header[1], sizeof(header[1]), 1, fp);
	if (ret != 1) // EOF reached
	{
	    badBytes++;
	    break;
	}
	if (verbose > 3) cerr << "byte[" << pos << "]=" 
			      << (unsigned int)header[1] << endl;
	pos++;
	if (!IOMpegFrameHeader::isSecondByte(header[0]))
	{
	    badBytes++;
	    continue;
	}
	    
	ret = fread(&header[2], sizeof(header[2]), 2, fp);
	if (ret != 2) // EOF reached
	{
	    badBytes++;
	    break;
	}
	if (verbose > 3) 
	{
	    cerr << "byte[" << pos << "]=" 
		 << (unsigned int)header[2] << endl;
	    cerr << "byte[" << pos+1 << "]=" 
		 << (unsigned int)header[3] << endl;
	}
	pos += 2;

	//-----------------------
	// analyze frame header:
	//-----------------------
	frameNum++;
	IOMpegFrameHeader frameHead(header);
	bool printFrameInfo = false;
	if ((verbose == 1) && !(frameNum % 1000))
	{
	    cerr << "[" << frameNum << "]\r";
	    cerr.flush();
	    printFrameInfo = true;
	}
	if (verbose > 2) 
	{
	    cerr << "Frame " << frameNum << ": " << frameHead.getInfo();
	}

	if (!frameHead.validHeader())
	{
	    badBytes += 2;
	    badFrames++;
	    if (verbose > 2)
		cerr << "skipping invalid frame " << frameNum << endl;
	    continue;
	}

	double curtime = frameHead.getFrameTime();
	if (0 < curtime) acctime += curtime;

	//------------------------
	// read CRC if available:
	//------------------------
	uint16 crc = 0;
	if (frameHead.hasCRC())
	{
	    if (verbose > 3) cerr << "read CRC" << endl;
	    ret = fread(&crc, sizeof(crc), 1, fp);
	    if (ret != 1) break; // EOF reached
	    if (verbose > 2)
	    {
		cerr << "byte[" << pos << "]=" << (crc & 0xff) << endl;
		cerr << "byte[" << pos+1 << "]=" << ((crc >> 8) & 0xff)
		     << endl;
	    }
	    pos += 2;
	    if (verbose > 3) cerr << "CRC: " << crc << endl;
	}

	//-------------------------------
	// read audio data if available:
	//-------------------------------
	int remSize = frameHead.getAudioSize();
	if (remSize > 0)
	{
	    if (verbose > 3) cerr << "read " << remSize 
				  << " bytes from frame" << endl;
	    uint8 frameRest[IOMpegFrameHeader::MAX_FRAME_SIZE];
	    ret = fread(frameRest, 1, remSize, fp);
	    if (ret != remSize) break; // EOF reached
	    pos += remSize;

	    if (printFrameInfo)
	    {
		cerr << inputFileName << ": [" << frameNum << "]";
		cerr.flush();
	    }
	    //---------------------------------
	    // write complete frame if needed:
	    //---------------------------------
	    int frameStat;
	    if (useTime) frameStat = inRange(acctime/1000, t1, t2);
	    else	     frameStat = inRange(frameNum, f1, f2);
	    if (1 == frameStat) // frame has passed range
	    {
		if (printFrameInfo)
		{
		    cerr << "\r";
		    cerr.flush();
		}
		break; // nothing more to do with this file
	    }
	    else if (0 == frameStat) // this is a frame to write
	    {
		frameOutNum++;
		if (printFrameInfo)
		{
		    cerr << " -> " << outputFileName << ": [" << frameOutNum << "]";
		}
		if (verbose > 2) cerr << "write frame " << frameNum 
				      << endl;
		// write header:
		ret = fwrite(header, sizeof(header), 1, out);
		if (ret != 1) break; // EOF reached
		
		// write CRC (if available):
		if (frameHead.hasCRC())
		{
		    ret = fwrite(&crc, sizeof(crc), 1, out);
		    if (ret != 1) break; // EOF reached
		}

		// write audio data:
		ret = fwrite(frameRest, 1, remSize, out);
		if (ret != remSize) break; // EOF reached
	    }
	    if (printFrameInfo)
	    {
		cerr << "\r";
		cerr.flush();
	    }
	}
    }
    if (verbose == 1) cerr << endl;
    if (verbose > 1) cerr << "close file: " << inputFileName << endl;
    (void)fclose(fp);
    fp = 0;
    if (verbose > 1)
    {
	cerr << "Statistics:" << endl;
	cerr << "Frames: " << frameNum 
	     << ", bad: " << badFrames << " Frames, bad Bytes: " 
	     << badBytes << " Bytes" << endl;
	cerr << "Time:   " << acctime << " ms = "; 
	double seconds = acctime/1000;
	int min = (int)seconds / 60;
	int sec = (int)seconds % 60;
	double ms = (seconds - (min * 60 + sec)) * 1000;
	char buf[80];
	(void)sprintf(buf,"%02d:%02d.%03.0f", min, sec, ms);
	cerr << buf << endl; 
    }
    else
    {
	if (badFrames)
	    cerr << "Warning: skipped " << badFrames << " bad Frames" 
		 << endl;
    }
} // end of handle_input_file()


//----------------------------------------------------------------
// read input files and write specified part into output file
// On Error: throws an exception
//----------------------------------------------------------------
static void process_files
(
    const FileSettings& files,     // what to do with the input files
    const string& outputFileName   // name of output file
)
{
    // open output file:
    if (verbose > 1) cerr << "open file: " << outputFileName << endl;
    FILE* out = stdout;
    if ("-" != outputFileName)
    {
	out = fopen(outputFileName.c_str(), "w");
	if (!out) 
	{
	    cerr << "Cannot open output file \"" << outputFileName << "\": " 
		 << strerror(errno) << endl;
	}
    }

    // handle all input files:
    unsigned int frameOutNum = 0;
    for (unsigned int i=0; i < files.getNum(); i++)
    {
	handle_input_file(files, i, outputFileName, out, frameOutNum);
    }
    if ("-" != outputFileName) (void)fclose(out);
    if (verbose == 1) cerr << endl;
} // end of process_files()


// --- global functions ---


int main(int argc, char* argv[])
{
    //------------------------------
    // handle command line options:
    //------------------------------
    FileSettings files;
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'i': files.addFileName(optarg);    break;
	case 'f': files.addFrameRange(optarg);  break;
	case 't': files.addTimeRange(optarg);   break;
	case 'v': verbose++;                    break;
	case 'h': usage(true);                  exit(0);
	case 'V': print_version();              exit(0);
	default : usage();                      exit(1);
	}
    }
    if (0 == files.getNum()) 
    {
	cerr << "Error: missing input file" << endl;
	usage();
	exit(1);
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing output file" << endl;
	usage();
	exit(1);
    }
    // the last argument is the output file name;
    string outputFileName(argv[optind++]);
    // check for too many arguments:
    if (optind < argc) 
    {
	cerr << "Warning: ignoring additional arguments behind \"" 
	     << outputFileName << "\"" << endl;
    }

    //--------------
    // check files:
    //--------------
    try { check_input_files(files); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    //--------------------------
    // process all input files:
    //--------------------------
    try { process_files(files, outputFileName); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
} // end of main()


// --- end of file ---
