//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/mpegtools/mpeginfo.c,v $
// $Id: mpeginfo.c,v 1.5 2004/04/21 21:43:12 ah Exp $ 
//
// Description: print information about MPEG files
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: mpeginfo.c,v 1.5 2004/04/21 21:43:12 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
#include <AH/Config/Types.h>
USING_NAMESPACE(AH)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/IO/Mpeg/FrameHeader.h>
USING_NAMESPACE(AH)

// --- local constants ---

static const char* cmdLineOptions = "hi:s:vV";
static const char* version = "1.1";

// --- local variables ---

static int verbose = 0;

// --- local classes ---

class FileSettings
{
 public:
    // add new file name
    void addFileName(const string& name)
    {
	fileName.push_back(name);
    }
    // get number of entries
    unsigned int getNum() const { return fileName.size(); }
    // get specified filename
    const string& getFileName(unsigned int i) const
    {
	if (i < fileName.size()) return fileName[i]; 
	return empty;
    }

 private:
    vector<string> fileName;
    string empty;
}; // class FileSettings


// --- local functions ---


static void usage(bool help = false)
{
    cerr << "Usage: mpeginfo [-" << cmdLineOptions << "] file [file, ...]\n"
	 << "       print information about MPEG file(s).\n"
	 << "  options:\n"
	 << "   -h                  - print this help message\n"
	 << "   -i <input file>     - specify input file\n"
	 << "   -v                  - verbose mode (increase debug level)\n"
	 << "   -V                  - print version\n"
	 << "  arguments:\n"
	 << "   <file>              - MPEG file (MPEG-1/2/2.5 Layer 1/2/3)\n"
	 << "\n"
	 << "All files must be MPEG files. The supported versions are:\n"
	 << "- MPEG-1 Layer 1, Layer 2 (MP2) and Layer 3 (MP3)\n" 
	 << "- MPEG-2 Layer 1, Layer 2 and Layer 3\n" 
	 << "- MPEG-2.5 Layer 1, Layer 2 and Layer 3\n" 
	 << endl;
    if (!help) return;

    cerr << "Examples:\n"
	 << "  mpeginfo a.mpg b.mpg c.mpg\n"
	 << "  - print information about the files a.mpg, b.mpg and c.mpg\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "mpeginfo Vs " << version << " " 
	 << "(c) 2000 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//----------------------------------------------------------------
// check input files for existence and for matching audio format
// On Error: throws an exception
//----------------------------------------------------------------
static void process_files
(
    const FileSettings& files
)
{
    int32 wrsize = 1;
    for (unsigned int i=0; (i < files.getNum()) && (0 < wrsize); i++)
    {
	const string& inputFileName = files.getFileName(i);

	// open input file:
	if (verbose > 1) cerr << "open file: " << inputFileName << endl;
	FILE* fp = fopen(inputFileName.c_str(), "r");
	if (!fp) continue;

	// read from input file:
	if (verbose > 2) cerr << "reading file: " << inputFileName << endl;
	unsigned int frameNum = 0;
	unsigned int pos = 0;
	unsigned int badBytes = 0;
	unsigned int badFrames = 0;
	double acctime = 0;
	while(1)
	{
	    uint8 header[4]; 

	    //--------------------
	    // read frame header:
	    //--------------------
	    int ret = fread(&header[0], sizeof(header[0]), 1, fp);
	    if (ret != 1) break; // EOF reached
	    if (verbose > 3) cerr << "byte[" << pos << "]=" 
			      << (unsigned int)header[0] << endl;
	    pos++;
	    if (!IOMpegFrameHeader::isFirstByte(header[0]))
	    {
		badBytes++;
		continue;
	    }
	    
	    ret = fread(&header[1], sizeof(header[1]), 1, fp);
	    if (ret != 1) // EOF reached
	    {
		badBytes++;
		break;
	    }
	    if (verbose > 3) cerr << "byte[" << pos << "]=" 
			      << (unsigned int)header[1] << endl;
	    pos++;
	    if (!IOMpegFrameHeader::isSecondByte(header[0]))
	    {
		badBytes++;
		continue;
	    }
	    
	    ret = fread(&header[2], sizeof(header[2]), 2, fp);
	    if (ret != 2) // EOF reached
	    {
		badBytes++;
		break;
	    }
	    if (verbose > 3) 
	    {
		cerr << "byte[" << pos << "]=" 
		     << (unsigned int)header[2] << endl;
		cerr << "byte[" << pos+1 << "]=" 
		     << (unsigned int)header[3] << endl;
	    }
	    pos += 2;

	    //-----------------------
	    // analyze frame header:
	    //-----------------------
	    frameNum++;
	    IOMpegFrameHeader frameHead(header);
	    if (verbose > 2) 
	    {
		cerr << "Frame " << frameNum << ": " << frameHead.getInfo();
	    }
	    if (!frameHead.validHeader())
	    {
		badBytes += 2;
		badFrames++;
		if (verbose > 1)
		    cerr << "skipping invalid frame " << frameNum << endl;
		continue;
	    }

	    double curtime = frameHead.getFrameTime();
	    if (0 < curtime) acctime += curtime;

	    //------------------------
	    // read CRC if available:
	    //------------------------
	    if (frameHead.hasCRC())
	    {
		if (verbose > 3) cerr << "read CRC: ";
		uint16 crc;
		ret = fread(&crc, sizeof(crc), 1, fp);
		if (ret != 1) break; // EOF reached
		if (verbose > 2)
		{
		    cerr << "byte[" << pos << "]=" << (crc & 0xff) << endl;
		    cerr << "byte[" << pos+1 << "]=" << ((crc >> 8) & 0xff)
			 << endl;
		}
		pos += 2;
		if (verbose > 3) cerr << crc << endl;
	    }

	    //-------------------------------
	    // read audio data if available:
	    //-------------------------------
	    int remSize = frameHead.getAudioSize();
	    if (remSize > 0)
	    {
		if (verbose > 3) cerr << "read " << remSize 
				  << " bytes from frame" << endl;
		uint8 frameRest[IOMpegFrameHeader::MAX_FRAME_SIZE];
		ret = fread(frameRest, 1, remSize, fp);
		if (ret != remSize) break; // EOF reached
		pos += remSize;
	    }

	}
	if (verbose > 1) cerr << "close file: " << inputFileName << endl;
	(void)fclose(fp);
	fp = 0;

	cout << "File    : " << inputFileName << endl;
	cout << "Frames  : " << frameNum 
	     << ", bad: " << badFrames << " Frames, "
	     << badBytes << " Bytes" << endl;
	cout << "Playtime:   " << acctime << " ms = "; 
	double seconds = acctime/1000;
	int min = (int)seconds / 60;
	int sec = (int)seconds % 60;
	double ms = (seconds - (min * 60 + sec)) * 1000;
	char buf[80];
	(void)sprintf(buf,"%02d:%02d.%03.0f\n", min, sec, ms);
	cout << buf << endl; 
    }
} // end of process_files()


// --- global functions ---


int main(int argc, char* argv[])
{
    //------------------------------
    // handle command line options:
    //------------------------------
    FileSettings files;
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'v': verbose++;                    break;
	case 'h': usage(true);                  exit(0);
	case 'V': print_version();              exit(0);
	default : usage();                      exit(1);
	}
    }
    // remaining arguments are input files:
    while (optind < argc) 
    {
	files.addFileName(argv[optind++]);
    }
    if (0 == files.getNum()) 
    {
	cerr << "Error: missing input file" << endl;
	usage();
	exit(1);
    }

    //--------------------------
    // process all input files:
    //--------------------------
    try { process_files(files); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
} // end of main()


// --- end of file ---
